''' SVG path elements '''

from __future__ import annotations
from typing import Union
from .config import config


def fmt(f: float) -> str:
    ''' String formatter, stripping trailing zeros '''
    p = f'.{config.precision}f'
    s = format(float(f), p)
    return s.rstrip('0').rstrip('.')  # Strip trailing zeros


class Point:
    ''' (X, Y) Point class '''
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return f'Point({self.x}, {self.y})'

    def __add__(self, p):
        return Point(self.x + p.x, self.y + p.y)

    def __sub__(self, p):
        return Point(self.x - p.x, self.y - p.y)

    def __truediv__(self, i):
        return Point(self.x / i, self.y / i)

    def xform(self, a: float, b: float, c: float, d: float,
              e: float, f: float, m: float, n: float) -> Point:
        ''' Apply Transform (for compound glyph) '''
        x = m * (a/m * self.x + c/m * self.y + e)
        y = n * (b/n * self.x + d/n * self.y + f)
        return Point(x, y)


class Moveto:
    ''' SVG Move To '''
    def __init__(self, p: Point):
        self.p = p

    def __repr__(self):
        return f'Moveto({self.p.x}, {self.p.y})'

    def path(self, x0: float = 0, y0: float = 0, scale: float = 1) -> str:
        ''' Get SVG path '''
        return f'M {fmt(x0 + self.p.x * scale)} {fmt(y0 - self.p.y * scale)} '

    def xform(self, a: float, b: float, c: float, d: float,
              e: float, f: float, m: float, n: float) -> Moveto:
        ''' Transform the SVG path '''
        return Moveto(self.p.xform(a, b, c, d, e, f, m, n))

    def ymin(self) -> float:
        ''' Minimum Y value '''
        return self.p.y

    def ymax(self) -> float:
        ''' Maximum Y value '''
        return self.p.y

    def xmin(self) -> float:
        ''' Minimum X value '''
        return self.p.x

    def xmax(self) -> float:
        ''' Maximum X value '''
        return self.p.x

    def points(self) -> tuple[list[Point], list[bool]]:
        ''' Get control points '''
        return [self.p], [False]  # [points], [control?]


class Lineto:
    ''' SVG Line To '''
    def __init__(self, p: Point):
        self.p = p

    def __repr__(self):
        return f'Lineto({self.p.x}, {self.p.y})'

    def path(self, x0: float = 0, y0: float = 0, scale: float = 1) -> str:
        ''' Get SVG path '''
        return f'L {fmt(x0 + self.p.x * scale)} {fmt(y0 - self.p.y * scale)} '

    def xform(self, a: float, b: float, c: float, d: float,
              e: float, f: float, m: float, n: float) -> Lineto:
        ''' Transform the SVG path '''
        return Lineto(self.p.xform(a, b, c, d, e, f, m, n))

    def ymin(self) -> float:
        ''' Minimum Y value '''
        return self.p.y

    def ymax(self) -> float:
        ''' Maximum Y value '''
        return self.p.y

    def xmin(self) -> float:
        ''' Minimum X value '''
        return self.p.x

    def xmax(self) -> float:
        ''' Maximum X value '''
        return self.p.x

    def points(self) -> tuple[list[Point], list[bool]]:
        ''' Get control points '''
        return [self.p], [False]  # [points], [control?]


class Quad:
    ''' SVG Quadratic Bezier.
        First point of Bezier should already be set in path.
        p1 is control points, p2 is endpoint.
    '''
    def __init__(self, p1: Point, p2: Point):
        self.p1 = p1
        self.p2 = p2

    def __repr__(self):
        return f'Quad({self.p1.x}, {self.p1.y}; {self.p2.x} {self.p2.y})'

    def path(self, x0: float = 0, y0: float = 0, scale: float = 1) -> str:
        ''' Get SVG path '''
        return (f'Q {fmt(x0 + self.p1.x * scale)} {fmt(y0-self.p1.y * scale)} '
                f'{fmt(x0+self.p2.x * scale)} {fmt(y0-self.p2.y * scale)} ')

    def xform(self, a: float, b: float, c: float, d: float,
              e: float, f: float, m: float, n: float) -> Quad:
        ''' Transform the SVG path '''
        return Quad(self.p1.xform(a, b, c, d, e, f, m, n),
                    self.p2.xform(a, b, c, d, e, f, m, n))

    def ymin(self) -> float:
        ''' Minimum Y value '''
        return min(self.p1.y, self.p2.y)

    def ymax(self) -> float:
        ''' Maximum Y value '''
        return max(self.p1.y, self.p2.y)

    def xmin(self) -> float:
        ''' Minimum X value '''
        return min(self.p1.x, self.p2.x)

    def xmax(self) -> float:
        ''' Maximum X value '''
        return max(self.p1.x, self.p2.x)

    def points(self) -> tuple[list[Point], list[bool]]:
        ''' Get control points '''
        return [self.p1, self.p2], [True, False]  # [points], [control?]


class Cubic:
    ''' SVG Cubic Bezier.
        First point of Bezier should already be set in path.
        p1, p2 are control points, p3 is endpoint.
    '''
    def __init__(self, p1: Point, p2: Point, p3: Point):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3

    def __repr__(self):
        return f'Cubic({self.p1.x}, {self.p1.y}; {self.p2.x}, {self.p2.y}; {self.p3.x} {self.p3.y})'

    def path(self, x0: float = 0, y0: float = 0, scale: float = 1) -> str:
        ''' Get SVG path '''
        return (f'C {fmt(x0+self.p1.x * scale)} {fmt(y0-self.p1.y * scale)} '
                f'{fmt(x0+self.p2.x * scale)} {fmt(y0-self.p2.y * scale)} '
                f'{fmt(x0+self.p3.x * scale)} {fmt(y0-self.p3.y * scale)} ')

    def xform(self, a: float, b: float, c: float, d: float,
              e: float, f: float, m: float, n: float) -> Cubic:
        ''' Transform the SVG path '''
        return Cubic(self.p1.xform(a, b, c, d, e, f, m, n),
                     self.p2.xform(a, b, c, d, e, f, m, n),
                     self.p3.xform(a, b, c, d, e, f, m, n))

    def ymin(self) -> float:
        ''' Minimum Y value '''
        return min(self.p1.y, self.p2.y, self.p3.y)

    def ymax(self) -> float:
        ''' Maximum Y value '''
        return max(self.p1.y, self.p2.y, self.p3.y)

    def xmin(self) -> float:
        ''' Minimum X value '''
        return min(self.p1.x, self.p2.x, self.p3.x)

    def xmax(self) -> float:
        ''' Maximum X value '''
        return max(self.p1.x, self.p2.x, self.p3.x)

    def points(self) -> tuple[list[Point], list[bool]]:
        ''' Get control points '''
        return [self.p1, self.p2, self.p3], [True, True, False]  # [points], [control?]


SVGOpType = Union[Moveto, Lineto, Quad, Cubic]



''' Read font file and write glyphs to SVG '''

from __future__ import annotations
from typing import Literal, Union, Optional, Sequence, Dict
import math
from pathlib import Path
from collections import namedtuple
from itertools import accumulate
import importlib.resources as pkg_resources
import xml.etree.ElementTree as ET

from .config import config
from .fontread import FontReader
from . import gpos
from . import gsub
from .cmap import Cmap12, Cmap4
from .glyph import SimpleGlyph, CompoundGlyph
from .glyphcff import read_glyph_cff, CFF
from .glyphglyf import read_glyph_glyf
from .findfont import find_font
from .fonttypes import (AdvanceWidth, Layout, Header, Table,
                        FontInfo, FontNames, Symbols)
from .svgpath import fmt


class Font:
    ''' Class to read/parse a OpenType/TTF and write glyphs to SVG

        Args:
            name: File name of the font. Defaults to DejaVuSans.
    '''
    def __init__(self, name: Optional[Union[str, Path]] = None,
                 searchpaths: Optional[Sequence[str | Path]] = None):
        self.fname = None
        if name:
            if Path(name).exists():
                self.fname = Path(name)
            else:
                self.fname = find_font(name, searchpaths)
                if self.fname is None:
                    raise FileNotFoundError(f'Font {name} not found.')
        else:
            with pkg_resources.path('ziafont.fonts', 'DejaVuSans.ttf') as p:
                self.fname = p

        with open(self.fname, 'rb') as f:
            self.fontfile = FontReader(f.read())

        self.tables: dict[str, Table] = {}
        self.features: dict[str, bool] = {}
        self.info = self._loadfont()  # Load in all the font metadata
        self._glyphs: Dict[int, Union[SimpleGlyph, CompoundGlyph]] = {}
        self._glyphids: Dict[str, int] = {}

    def _loadfont(self) -> FontInfo:
        ''' Read font metadata '''
        self._readtables()
        header, layout = self._readheader()
        self.advwidths = self._readwidths(header.numlonghormetrics)
        self._readcmap()
        names = self._readnames()
        info = FontInfo(self.fname, names, header, layout)

        self.gpos = None
        if 'GPOS' in self.tables:
            self.gpos = gpos.Gpos(self.tables['GPOS'].offset, self.fontfile)
            self.features.update(self.gpos.init_user_features())

        self.gsub = None
        if 'GSUB' in self.tables:
            self.gsub = gsub.Gsub(self.tables['GSUB'].offset, self.fontfile)
            self.features.update(self.gsub.init_user_features())

        self.cffdata: Optional[CFF] = None
        if 'CFF ' in self.tables:
            self.cffdata = CFF(self)

        return info

    def _readtables(self) -> None:
        ''' Read list of tables in the font, and verify checksums '''
        self.fontfile.seek(0)
        scalartype = self.fontfile.readuint32()
        numtables = self.fontfile.readuint16()
        searchrange = self.fontfile.readuint16()
        entryselector = self.fontfile.readuint16()
        rangeshift = self.fontfile.readuint16()   # numtables*16-searchrange

        # Table Directory (table 5)
        self.tables = {}
        for i in range(numtables):
            tag = self.fontfile.read(4).decode()
            self.tables[tag] = Table(checksum=self.fontfile.readuint32(),
                                     offset=self.fontfile.readuint32(),
                                     length=self.fontfile.readuint32())

        if 'glyf' not in self.tables and 'CFF ' not in self.tables:
            raise ValueError('Unsupported font (no glyf or CFF table).')

    def verifychecksum(self) -> None:
        ''' Verify checksum of all font tables. Raises ValueError if invalid '''
        for table in self.tables.keys():
            if table != 'head':
                self._verifychecksum(table)

    def _verifychecksum(self, table: str) -> None:
        ''' Verify checksum of table. Raises ValueError if invalid. '''
        tb = self.tables[table]
        self.fontfile.seek(tb.offset)
        s = 0
        nlongs = (tb.length + 3) // 4
        s = sum(self.fontfile.readuint32() for i in range(nlongs)) & 0xffffffff
        if s != tb.checksum:
            raise ValueError(f'Table {table} checksum {s} != saved checksum {tb.checksum}')

    def _readheader(self) -> tuple[Header, Layout]:
        ''' Read Font "head" and "hhea" tables '''
        version = self.fontfile.readuint32(self.tables['head'].offset)
        revision = self.fontfile.readuint32()
        chksumadjust = self.fontfile.readuint32()
        magic = self.fontfile.readuint32()
        assert magic == 0x5f0f3cf5
        flags = self.fontfile.readuint16()
        unitsperem = self.fontfile.readuint16()
        created = self.fontfile.readdate()
        modified = self.fontfile.readdate()
        xmin = self.fontfile.readint16()
        ymin = self.fontfile.readint16()
        xmax = self.fontfile.readint16()
        ymax = self.fontfile.readint16()
        macstyle = self.fontfile.readuint16()
        lowestrecppem = self.fontfile.readuint16()
        directionhint = self.fontfile.readint16()
        indextolocformat = self.fontfile.readint16()
        glyphdataformat = self.fontfile.readint16()

        # hhea table with other parameters
        _ = self.fontfile.readuint32(self.tables['hhea'].offset)  # version
        ascent = self.fontfile.readint16()
        descent = self.fontfile.readint16()
        linegap = self.fontfile.readint16()
        advwidthmax = self.fontfile.readuint16()
        minleftbearing = self.fontfile.readint16()
        minrightbearing = self.fontfile.readint16()
        xmaxextent = self.fontfile.readint16()
        caretsloperise = self.fontfile.readint16()
        caretsloperun = self.fontfile.readint16()
        caretoffset = self.fontfile.readint16()
        for i in range(4):
            self.fontfile.readint16()  # Skip reserved
        metricformat = self.fontfile.readint16()
        numlonghormetrics = self.fontfile.readuint16()
        advwidth = AdvanceWidth(advwidthmax, minleftbearing)

        self.fontfile.readuint32(self.tables['maxp'].offset)
        numglyphs = self.fontfile.readuint16()

        layout = Layout(unitsperem, xmin, xmax, ymin, ymax, ascent, descent,
                        advwidth, minleftbearing, minrightbearing)
        header = Header(version, revision, chksumadjust, magic, flags,
                        created, modified, macstyle,
                        lowestrecppem, directionhint, indextolocformat,
                        glyphdataformat, numlonghormetrics, numglyphs)
        return header, layout

    def _readnames(self) -> FontNames:
        ''' Read the "name" table which includes the font name,
            copyright, and other info
        '''
        namefmt = self.fontfile.readuint16(self.tables['name'].offset)
        nameids = [''] * 15  # Empty strings for nameId table

        if namefmt == 0:  # '1' is not supported
            count = self.fontfile.readuint16()
            strofst = self.fontfile.readuint16()
            namerecords = []
            for i in range(count):
                platformId = self.fontfile.readuint16()
                platformSpecificId = self.fontfile.readuint16()
                languageId = self.fontfile.readuint16()
                nameId = self.fontfile.readuint16()
                length = self.fontfile.readuint16()
                offset = self.fontfile.readuint16()
                namerecords.append((platformId, platformSpecificId, languageId,
                                    nameId, length, offset))
            for record in namerecords:
                self.fontfile.seek(self.tables['name'].offset + strofst + record[5])
                name = self.fontfile.read(record[4])
                if record[3] < 16:
                    if record[0] in [0, 3]:  # Microsoft and Unicode formats
                        nameids[record[3]] = name.decode('utf-16be')

        return FontNames(*nameids)

    def _readwidths(self, numlonghormetrics: int) -> list[AdvanceWidth]:
        ''' Read `advanceWidth` and `leftsidebearing` from "htmx" table '''
        self.fontfile.seek(self.tables['hmtx'].offset)
        advwidths = []
        for i in range(numlonghormetrics):
            w = self.fontfile.readuint16()
            b = self.fontfile.readint16()
            advwidths.append(AdvanceWidth(w, b))
        return advwidths

    def _readcmap(self) -> None:
        ''' Read "cmap" table and select a cmap for locating glyphs from characters.
            Cmap formats 4 and 12 are supported.
        '''
        platforms = {0: 'unicode', 1: 'macintosh', 3: 'windows'}
        version = self.fontfile.readint16(self.tables['cmap'].offset)
        numtables = self.fontfile.readint16()
        CMapTable = namedtuple('CMapTable', ['platform', 'platformid', 'offset'])
        cmaptables = []
        for i in range(numtables):
            cmaptables.append(CMapTable(
                platforms.get(self.fontfile.readuint16()),
                self.fontfile.readuint16(),
                self.fontfile.readuint32()))

        self.cmap: Optional[Union[Cmap12, Cmap4]] = None  # Active cmap
        self.cmaps: list[Union[Cmap12, Cmap4]] = []
        cmap: Union[Cmap12, Cmap4]
        for ctable in cmaptables:
            ctable_ofst = self.tables['cmap'].offset + ctable.offset
            cmapformat = self.fontfile.readuint16(ctable_ofst)
            if cmapformat == 4:
                endcodes = []
                startcodes = []
                iddeltas = []
                idrangeoffset = []
                glyphidxarray = []
                length = self.fontfile.readuint16()
                lang = self.fontfile.readuint16()
                segcount = self.fontfile.readuint16() // 2
                searchrange = self.fontfile.readuint16()
                entryselector = self.fontfile.readuint16()
                rangeshift = self.fontfile.readuint16()
                for i in range(segcount):
                    endcodes.append(self.fontfile.readuint16())
                _ = self.fontfile.readuint16()  # reserved pad
                for i in range(segcount):
                    startcodes.append(self.fontfile.readuint16())
                for i in range(segcount):
                    iddeltas.append(self.fontfile.readuint16())
                for i in range(segcount):
                    idrangeoffset.append(self.fontfile.readuint16())

                # Length of glyph array comes from total length of cmap table
                # //2 because len is in bytes, but table glyphidxarray is 16-bit
                glyphtablelen = (length - (self.fontfile.tell() - ctable_ofst)) // 2
                for i in range(glyphtablelen):
                    glyphidxarray.append(self.fontfile.readuint16())
                cmap = Cmap4(ctable.platform, ctable.platformid,
                             startcodes, endcodes, idrangeoffset, iddeltas, glyphidxarray)
                if self.cmap is None:
                    self.cmap = cmap
                self.cmaps.append(cmap)

            elif cmapformat == 12:
                _ = self.fontfile.readuint16()
                length = self.fontfile.readuint32()
                lang = self.fontfile.readuint32()
                ngroups = self.fontfile.readuint32()
                starts = []
                ends = []
                glyphstarts = []
                for i in range(ngroups):
                    starts.append(self.fontfile.readuint32())
                    ends.append(self.fontfile.readuint32())
                    glyphstarts.append(self.fontfile.readuint32())
                cmap = Cmap12(ctable.platform, ctable.platformid, starts, ends, glyphstarts)
                if self.cmap is None or isinstance(self.cmap, Cmap4):
                    self.cmap = cmap
                self.cmaps.append(cmap)

        if len(self.cmaps) == 0:
            raise ValueError('No suitable cmap table found in font.')

    def usecmap(self, cmapidx: int) -> None:
        ''' Select cmap table by index. Only supported tables are included. '''
        self.cmap = self.cmaps[cmapidx]

    def scripts(self):
        ''' Get list of scripts in the font '''
        scripts = set()
        if self.gpos:
            scripts = scripts.union(set(self.gpos.scripts.keys()))
        if self.gsub:
            scripts = scripts.union(set(self.gsub.scripts.keys()))
        return list(scripts)

    def languages(self, script='DFLT'):
        ''' Get list of languages in the script '''
        langs = set()
        if self.gpos:
            s = self.gpos.scripts.get(script, None)
            if s is not None:
                langs = langs.union(set(s.languages.keys()))

        if self.gsub:
            s = self.gsub.scripts.get(script, None)
            if s is not None:
                langs = langs.union(set(s.languages.keys()))
        return list(langs)

    def language(self, script, language):
        ''' Set script/language to use '''
        scripts = self.scripts()
        if script == 'DFLT' and 'DFLT' not in scripts:
            script = 'latn' if 'latn' in scripts else scripts[0]

        if script not in self.scripts():
            raise ValueError(f'Script {script} not defined in font')

        if language not in self.languages(script):
            raise ValueError(f'Language {language} not defined in font')

        self.features = {}
        if self.gpos:
            self.gpos.language.script = script
            self.gpos.language.language = language
            self.features.update(self.gpos.init_user_features())
        if self.gsub:
            self.gsub.language.script = script
            self.gsub.language.language = language
            self.features.update(self.gsub.init_user_features())

    def glyphindex(self, char: str) -> int:
        ''' Get index of character glyph '''
        gid = self._glyphids.get(char)
        if gid is None:
            gid = self.cmap.glyphid(char)  # type: ignore
            self._glyphids[char] = gid
        return gid

    def glyph(self, char: str) -> SimpleGlyph:
        ''' Get the Glyph for the character '''
        index = self.glyphindex(char)        # Glyph Number
        return self.glyph_fromid(index)

    def glyph_fromid(self, glyphid: int) -> Union[SimpleGlyph, CompoundGlyph]:
        ''' Read a glyph from the "glyf" table

            Args:
                glyphid: Glyph index used to find glyph data
        '''
        glyph = self._glyphs.get(glyphid)
        if glyph is None:
            if 'CFF ' in self.tables:
                glyph = read_glyph_cff(glyphid, self)
            else:  # 'glyf' table
                glyph = read_glyph_glyf(glyphid, self)
            self._glyphs[glyphid] = glyph
        return glyph

    def advance(self, glyph: int, glyph2: int|None) -> int:
        ''' Get advance width in font units '''
        try:
            adv = self.advwidths[glyph].width
        except IndexError:
            adv = self.info.layout.advwidthmax.width

        if self.features.get('kern', True) and glyph2 and self.gpos:
            # Only getting x-advance for first glyph.
            pos = self.gpos.position(
                [self.glyph_fromid(glyph), self.glyph_fromid(glyph2)],
                {'kern': True})
            adv = pos[1][0]

        return adv

    def getsize(self, s) -> tuple[float, float]:
        ''' Calculate width and height (including ascent/descent) of string '''
        txt = Text(s, self)
        return txt.getsize()

    def text(self, s: str,
             size: Optional[float] = None,
             linespacing: float = 1,
             halign: Literal['left', 'center', 'right'] = 'left',
             valign: Literal['base', 'center', 'top'] = 'base',
             color: Optional[str] = None,
             rotation: float = 0,
             rotation_mode: str = 'anchor'):
        ''' Create a Text object using this font

            Args:
                s: String to convert.
                size: Font size in points
                linespacing: Space between lines
                halign: Horizontal Alignment
                valign: Vertical Alignment
                color: Color for string
                rotation: Rotation angle in degrees
                rotation_mode: Either 'default' or 'anchor', to
                    mimic Matplotlib behavoir. See:
                    https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
                kern: Use font kerning adjustment
        '''
        txt = Text(s, self, size=size, linespacing=linespacing, halign=halign, valign=valign,
                   color=color, rotation=rotation, rotation_mode=rotation_mode)
        return txt


class Text:
    ''' Draw text as SVG paths

        Args:
            s: String to draw
            font: Font name or ziafont.Font to use
            size: Font size in points
            linespacing: Spacing between lines
            color: Color for string
            halign: Horizontal Alignment
            valign: Vertical Alignment
            rotation: Rotation angle in degrees
            rotation_mode: Either 'default' or 'anchor', to
                mimic Matplotlib behavoir. See:
                https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
    '''
    def __init__(self, s: Union[str, Sequence[int]],
                 font: Optional[Union[str, Font]] = None,
                 size: Optional[float] = None,
                 linespacing: float = 1,
                 halign: Literal['left', 'center', 'right'] = 'left',
                 valign: Literal['base', 'center', 'top'] = 'base',
                 color: Optional[str] = None,
                 rotation: float = 0,
                 rotation_mode: str = 'anchor'):
        self.str = s
        self.halign = halign
        self.valign = valign
        self.color = color
        self.rotation = rotation
        self.rotation_mode = rotation_mode
        self.size = size if size else config.fontsize
        self.linespacing = linespacing
        if font is None or isinstance(font, str):
            if font in loadedfonts:
                self.font = loadedfonts[font]
            else:
                # Load the font and cache it for later
                self.font = Font(font)
                loadedfonts[str(font)] = self.font
        else:
            self.font = font
        self._symbols = self._buildstring()

    def svgxml(self) -> ET.Element:
        ''' Get SVG XML element '''
        svg = ET.Element('svg')
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        if isinstance(self.str, str):
            title = ET.SubElement(svg, 'title')
            title.text = self.str
        ret, (xmin, xmax, ymin, ymax) = self._drawon(svg)
        w = xmax-xmin
        h = ymax-ymin
        svg.attrib['width'] = fmt(w)
        svg.attrib['height'] = fmt(h)
        svg.attrib['viewBox'] = f'{fmt(xmin)} {fmt(ymin)} {fmt(w)} {fmt(h)}'
        return ret

    def svg(self) -> str:
        ''' Get SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def _repr_svg_(self):
        ''' Jupyter representer '''
        return self.svg()

    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0):
        ''' Draw text on the SVG '''
        svg, _ = self._drawon(svg, x, y)
        return svg

    def _drawon(self, svg: ET.Element, x: float = 0, y: float = 0):
        ''' Draw text on the SVG '''
        word, symbols, width, xmin, ymin, ymax = self._symbols
        height = ymax-ymin
        # Adjust vertical alignment
        yofst = {'base': 0,
                 'bottom': -ymax,
                 'top': -ymin,
                 'center': -height/2-ymin}.get(self.valign, 0)
        xofst = {'center': -width/2,
                 'right': -width}.get(self.halign, 0)
        xy = x + xofst, y + yofst

        # Get existing symbol/glyphs, add ones not there yet
        if config.svg2:
            existingsymbols = svg.findall('symbol')
            symids = [sym.attrib.get('id') for sym in existingsymbols]
            for sym in symbols:
                if sym not in symids:
                    svg.append(sym)

        xform = ''
        if xy != (0, 0):
            xform = f'translate({fmt(xy[0])} {fmt(xy[1])})'

        svg.append(word)

        if config.debug:  # Test viewbox
            rect = ET.SubElement(word, 'rect')
            rect.attrib['x'] = '0'
            rect.attrib['y'] = fmt(ymin)
            rect.attrib['width'] = fmt(width)
            rect.attrib['height'] = fmt(height)
            rect.attrib['fill'] = 'none'
            rect.attrib['stroke'] = 'red'
            circ = ET.SubElement(word, 'circle')
            circ.attrib['cx'] = fmt(-xofst)
            circ.attrib['cy'] = fmt(-yofst)
            circ.attrib['r'] = '3'
            circ.attrib['fill'] = 'red'
            circ.attrib['stroke'] = 'red'
        bbox = (xy[0]+xmin, xy[0]+width, xy[1]+ymin, xy[1]+ymax)

        if self.rotation:
            centerx = xy[0]  # Center of rotation
            centery = xy[1]
            costh = math.cos(math.radians(self.rotation))
            sinth = math.sin(math.radians(self.rotation))
            p1 = (xofst, ymin+yofst)  # Corners relative to rotation point
            p2 = (xofst+width, ymin+yofst)
            p3 = (xofst+width, ymax+yofst)
            p4 = (xofst, ymax+yofst)
            x1 = centerx + (p1[0]*costh + p1[1]*sinth) - xofst
            x2 = centerx + (p2[0]*costh + p2[1]*sinth) - xofst
            x3 = centerx + (p3[0]*costh + p3[1]*sinth) - xofst
            x4 = centerx + (p4[0]*costh + p4[1]*sinth) - xofst
            y1 = centery - (p1[0]*sinth - p1[1]*costh) - yofst
            y2 = centery - (p2[0]*sinth - p2[1]*costh) - yofst
            y3 = centery - (p3[0]*sinth - p3[1]*costh) - yofst
            y4 = centery - (p4[0]*sinth - p4[1]*costh) - yofst
            bbox = (min(x1, x2, x3, x4), max(x1, x2, x3, x4),
                    min(y1, y2, y3, y4), max(y1, y2, y3, y4))

            if self.rotation_mode == 'default':
                dx = {'left': x - bbox[0],
                      'right': x - bbox[1],
                      'center': x - (bbox[1]+bbox[0])/2}.get(self.halign, 0)
                dy = {'top': y - bbox[2],
                      'bottom': y - bbox[3],
                      'base': -sinth*dx,
                      'center': y - (bbox[3]+bbox[2])/2}.get(self.valign, 0)
                xform = f'translate({xy[0]+dx} {xy[1]+dy})'
                bbox = (bbox[0]+dx, bbox[1]+dx,
                        bbox[2]+dy, bbox[3]+dy)
            xform += f' rotate({-self.rotation} {-xofst} {-yofst})'

        if config.debug:
            rect = ET.SubElement(svg, 'rect')
            rect.attrib['x'] = fmt(bbox[0])
            rect.attrib['y'] = fmt(bbox[2])
            rect.attrib['width'] = fmt(bbox[1]-bbox[0])
            rect.attrib['height'] = fmt(bbox[3]-bbox[2])
            rect.attrib['fill'] = 'none'
            rect.attrib['stroke'] = 'blue'

        word.attrib['transform'] = xform
        if self.color:
            word.attrib['fill'] = self.color
        return svg, bbox

    def str_to_gids(self):
        if isinstance(self.str, str):
            lines = self.str.splitlines()
            gidlines = [[self.font.glyphindex(c) for c in line] for line in lines]
        else:
            gidlines = [self.str]  # Single line
        return gidlines

    def _buildstring(self) -> Symbols:
        ''' Create symbols and svg word in a <g> group tag, for placing in an svg '''
        scale = self.size / self.font.info.layout.unitsperem
        lineheight = self.size * self.linespacing

        gidlines = self.str_to_gids()
        yvals = [i*lineheight for i in range(len(gidlines))]  # valign == 'base'

        symbols: list[ET.Element] = []  # <symbol> elements
        linewidths: list[float] = []
        allglyphs: list[list[tuple[SimpleGlyph, int, int]]] = []  # (glyph, x, y) where x is left aligned, y is delta from baseline
        xmin = 0
        ymin = math.inf
        ymax = -math.inf
        for lineidx, glyphids in enumerate(gidlines):
            lineglyphs: list[tuple[SimpleGlyph, int, int]] = []

            # Apply glyph substitutions (GSUB)
            if self.font.gsub:
                glyphids = self.font.gsub.sub(glyphids, self.font.features)

            glyphs = [self.font.glyph_fromid(gid) for gid in glyphids]

            # Apply Glyph Positioning (GPOS)
            if self.font.gpos:
                xy = self.font.gpos.position(glyphs, self.font.features)
            else:
                xs: list[int] = list(accumulate([glyph.advance() for glyph in glyphs], initial=0))
                xy = [(x, 0) for x in xs[:-1]]

            lineglyphs = [(glyph, x*scale, -y*scale) for glyph, (x, y) in zip(glyphs, xy)]

            # Accumulate multi-line widths/heights
            if glyphs:
                xmin = min(xmin, glyphs[0].bbox.xmin*scale)
                last_advance = glyphs[-1].advance()
                linewidth = (xy[-1][0] + last_advance) * scale
                if glyphs[-1].bbox.xmax > last_advance:
                    # Make linewidth a bit wider to grab right edge
                    # that extends beyond advance width
                    linewidth += (glyphs[-1].bbox.xmax - last_advance) * scale
                    linewidth += (-glyphs[0].bbox.xmin) * scale
            else:
                # Blank line
                linewidth = 0
            linewidths.append(linewidth)
            allglyphs.append(lineglyphs)

            # Accumulate SVG symbols for each glyph
            symbols.extend([glyph.svgsymbol()
                            for glyph in glyphs if glyph.id not in [s.attrib['id'] for s in symbols]])

        # Place the glyphs based on halign
        word = ET.Element('g')
        totwidth = max(linewidths)
        for lineidx, (lineglyphs, linewidth) in enumerate(zip(allglyphs, linewidths)):
            if self.halign == 'center':
                leftshift = (totwidth - linewidth)/2
            elif self.halign == 'right':
                leftshift = totwidth - linewidth
            else:  # halign = 'left'
                leftshift = 0
            for glyph, x, dy in lineglyphs:
                elm = glyph.place(x+leftshift, yvals[lineidx]+dy, self.size)
                if elm is not None:
                    word.append(elm)

        ymax = yvals[-1] - min(glyph[0].bbox.ymin*scale - glyph[2] for glyph in allglyphs[-1])
        ymin = yvals[0] - max(glyph[0].bbox.ymax*scale - glyph[2] for glyph in allglyphs[0])

        if not config.svg2:
            symbols = []
        return Symbols(word, symbols, totwidth, xmin, ymin, ymax)

    def getsize(self) -> tuple[float, float]:
        ''' Calculate width and height (including ascent/descent) of string '''
        return self._symbols.width, self._symbols.ymax-self._symbols.ymin

    def getyofst(self) -> float:
        ''' Y-shift from bottom of bbox to 0 '''
        return -self._symbols.ymax


loadedfonts: Dict[str, Font] = {}



''' Common Font File Tables '''

from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from dataclasses import dataclass
from collections import namedtuple

if TYPE_CHECKING:
    from .fontread import FontReader


class Coverage:
    ''' Coverage Table - defines which glyphs apply to a lookup

        Parameters
        ----------
        ofst: Byte offset from start of file
        fontfile: Font file to read from
        nulltable: Set true if this table is null
    '''
    def __init__(self, ofst: int, fontfile: FontReader, nulltable: bool = False):
        self.ofst = ofst
        self.fontfile = fontfile
        self.nulltable = nulltable
        if not self.nulltable:
            fileptr = self.fontfile.tell()

            self.fontfile.seek(self.ofst)
            self.format = self.fontfile.readuint16()

            if self.format == 1:
                glyphcnt = self.fontfile.readuint16()
                self.glyphs = []
                for i in range(glyphcnt):
                    self.glyphs.append(self.fontfile.readuint16())

            elif self.format == 2:
                rangecnt = self.fontfile.readuint16()
                Range = namedtuple('Range', ['startglyph', 'endglyph', 'covidx'])
                self.ranges: list[Range] = []
                for i in range(rangecnt):
                    self.ranges.append(Range(
                        self.fontfile.readuint16(),
                        self.fontfile.readuint16(),
                        self.fontfile.readuint16()))
            else:
                raise ValueError(f'Bad coverage table format {self.format}')

            self.fontfile.seek(fileptr)  # Put file pointer back

    def covidx(self, glyph: int) -> Optional[int]:
        ''' Get coverage index for this glyph, or None if not in the coverage range '''
        if self.nulltable:
            return None

        if self.format == 1:
            try:
                idx: Optional[int] = self.glyphs.index(glyph)
            except (ValueError, IndexError):
                idx = None
        else:
            for r in self.ranges:
                if r.startglyph <= glyph <= r.endglyph:
                    idx = r.covidx + (glyph - r.startglyph)
                    break
            else:
                idx = None
        return idx

    def list_glyphs(self) -> list[int]:
        ''' List all glyph indexes covered by the table '''
        if self.format == 1:
            return [g for g in self.glyphs]  # make a copy

        glyphs = []
        for rng in self.ranges:
            glyphs.extend(list(range(rng.startglyph, rng.endglyph+1)))
        return glyphs

    def __repr__(self):
        return f'<CoverageTable {hex(self.ofst)}>'


class ClassDef:
    ''' Class Definition Table - defines a "class" of glyphs. '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.format = self.fontfile.readuint16()

        if self.format == 1:
            self.startglyph = self.fontfile.readuint16()
            glyphcnt = self.fontfile.readuint16()
            self.classvalues = []
            for i in range(glyphcnt):
                self.classvalues.append(self.fontfile.readuint16())

        elif self.format == 2:
            ClassRange = namedtuple('ClassRange', ['startglyph', 'endglyph', 'cls'])
            rangecnt = self.fontfile.readuint16()
            self.ranges = []
            for i in range(rangecnt):
                self.ranges.append(ClassRange(
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16()))
        else:
            raise ValueError

        self.fontfile.seek(fileptr)  # Put file pointer back

    def get_class(self, glyphid: int) -> int:
        ''' Get class by glyph id '''
        if self.format == 1:
            if self.startglyph <= glyphid < self.startglyph + len(self.classvalues):
                return self.classvalues[glyphid - self.startglyph]
            else:
                return 0

        else:
            for i, rng in enumerate(self.ranges):
                if rng.startglyph <= glyphid <= rng.endglyph:
                    return rng.cls
            return 0

    def __repr__(self):
        return f'<ClassDefTable {hex(self.ofst)}>'


class ClassDefBlank(ClassDef):
    ''' Blank class definition '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

    def get_class(self, glyphid: int) -> int:
        return 0


class Feature:
    ''' Feature Table for GPOS/GSUB '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.featureparamofst = self.fontfile.readuint16()
        lookupcnt = self.fontfile.readuint16()
        self.lookupids = []
        for i in range(lookupcnt):
            self.lookupids.append(self.fontfile.readuint16())

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<Feature {self.tag}, {hex(self.ofst)}>'


class Script:
    ''' GPOS/GSUB Script table '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.defaultLangSysOfst = self.fontfile.readuint16()
        self.languages = {}

        if self.defaultLangSysOfst:
            self.languages[''] = LanguageSystem(
                '',
                self.defaultLangSysOfst + self.ofst,
                self.fontfile)

        langsyscnt = self.fontfile.readuint16()
        for i in range(langsyscnt):
            tag = self.fontfile.read(4).decode()
            self.languages[tag] = (LanguageSystem(
                tag,
                self.fontfile.readuint16() + self.ofst,
                self.fontfile))

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<Script {self.tag}, {hex(self.ofst)}>'


class LanguageSystem:
    ''' GPOS/GSUB Language System Table '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.fontfile.readuint16()  # lookupordeoffset (reserved)
        self.reqdFeatureIndex = self.fontfile.readuint16()
        featurecount = self.fontfile.readuint16()
        self.featureidxs = []
        for i in range(featurecount):
            self.featureidxs.append(self.fontfile.readuint16())

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<LanguageSystem {self.tag}, {hex(self.ofst)}>'


@dataclass
class Language:
    ''' Font script and language '''
    script: str = 'latn'
    language: str = ''  # Default



''' Glyph classes '''

from __future__ import annotations
from typing import Union, Optional, TYPE_CHECKING

from .glyph import SimpleGlyph, CompoundGlyph, EmptyGlyph
from .fonttypes import GlyphComp, BBox, Xform
from .svgpath import Moveto, Lineto, Quad, Point, SVGOpType

if TYPE_CHECKING:
    from .font import Font


def glyphoffset(font, index: int) -> Optional[int]:
    ''' Get offset (from beginning of file) of glyph,
        Return None if no glyph (empty/space) at this index.
    '''
    if font.info.header.indextolocformat == 1:
        offset = font.fontfile.readuint32(font.tables['loca'].offset + index * 4)
        nextofst = font.fontfile.readuint32()
    else:
        offset = font.fontfile.readuint16(font.tables['loca'].offset + index * 2) * 2
        nextofst = font.fontfile.readuint16() * 2

    if offset == nextofst:
        # Empty glyphs (ie space) have no length.
        return None
    else:
        return offset + font.tables['glyf'].offset


def read_glyph_glyf(glyphid: int, font: Font) -> SimpleGlyph:
    ''' Read a glyph from the glyf table. '''
    offset = glyphoffset(font, glyphid)

    if offset is None:
        return EmptyGlyph(glyphid, font)

    if offset >= font.tables['glyf'].offset + font.tables['glyf'].length:
        return EmptyGlyph(glyphid, font)

    assert offset >= font.tables['glyf'].offset
    assert offset < font.tables['glyf'].offset + font.tables['glyf'].length

    font.fontfile.seek(offset)

    numcontours = font.fontfile.readint16()
    xmin = font.fontfile.readint16()
    ymin = font.fontfile.readint16()
    xmax = font.fontfile.readint16()
    ymax = font.fontfile.readint16()
    charbox = BBox(xmin, xmax, ymin, ymax)
    glyph: Union[SimpleGlyph, CompoundGlyph]
    if numcontours == -1:
        glyph = read_compoundglyph(font, glyphid, charbox)
    else:
        glyph = read_simpleglyph(font, glyphid, numcontours, charbox)
    return glyph


def read_simpleglyph(font: Font, index: int, numcontours: int, charbox: BBox) -> SimpleGlyph:
    ''' Read a symple glyph from the fontfile. Assumes file pointer is set '''
    fontfile = font.fontfile
    ends = []
    for i in range(numcontours):
        ends.append(font.fontfile.readuint16())
    instlength = fontfile.readuint16()
    fontfile.seek(instlength + fontfile.tell())  # Skip instructions

    numpoints = max(ends) + 1

    # flags
    ONCURVE = 0x01
    XSHORT = 0x02
    YSHORT = 0x04
    REPEAT = 0x08
    XDUP = 0x10
    YDUP = 0x20

    flags = []
    i = 0
    while i < numpoints:
        flag = fontfile.readuint8()
        if (flag & REPEAT):
            nrepeats = fontfile.readuint8() + 1  # Include the first one too
        else:
            nrepeats = 1
        flags.extend([flag] * nrepeats)
        i += nrepeats

    ctvals = [(f & ONCURVE) == 0 for f in flags]  # True for control point, False for real point
    xvals = []
    xval = 0  # Points are stored as deltas. Add them up as we go to get real points.
    for flag in flags:
        if (flag & XSHORT):  # X is one-byte
            if (flag & XDUP):
                xval += fontfile.readuint8()
            else:
                xval -= fontfile.readuint8()
        elif not (flag & XDUP):
            xval += fontfile.readint16()
        # else: xval stays the same
        xvals.append(xval)

    yvals = []
    yval = 0  # Add up deltas
    for flag in flags:
        if (flag & YSHORT):  # Y is one-byte
            if (flag & YDUP):
                yval += fontfile.readuint8()
            else:
                yval -= fontfile.readuint8()
        elif not (flag & YDUP):
            yval += fontfile.readint16()
        # else: yval stays the same
        yvals.append(yval)

    points = []
    controls = []
    start = 0
    for end in ends:
        stop = end + 1
        points.append([Point(x, y) for x, y in zip(xvals[start:stop], yvals[start:stop])])
        controls.append(ctvals[start:stop])
        start = stop

    operators: list[SVGOpType] = []
    for p, c in zip(points, controls):
        npoints = len(p)
        if c[0]:
            # Path STARTS with a control point. Wrap last point in path.
            # Unna-Regular.ttf is an example.
            if c[-1]:
                pim = (p[0] + p[-1])/2
                operators.append(Moveto(pim))  # Last point is also control
            else:
                operators.append(Moveto(p[-1]))
        else:
            operators.append(Moveto(p[0]))

        i = 0
        while i < npoints:
            if c[i]:  # This is a control point
                if i == npoints-1:
                    # Last point is control. End point wraps to start point.
                    if c[0]:  # First point is also control, use midpoint
                        pim = (p[0] + p[-1])/2
                        operators.append(Quad(p[i], pim))
                    else:  # First point is real
                        operators.append(Quad(p[i], p[0]))
                    i += 1
                elif c[i+1]:
                    # Next point is also control.
                    # End of this bezier is implied between two controls
                    pim = (p[i] + p[i+1])/2
                    operators.append(Quad(p[i], pim))
                    i += 1
                else:
                    # Next point is real. It's the endpoint of bezier
                    operators.append(Quad(p[i], p[i+1]))
                    i += 2
            else:
                operators.append(Lineto(p[i]))
                i += 1

    glyph = SimpleGlyph(index, operators, charbox, font)
    return glyph


def read_compoundglyph(font: Font, index: int, charbox: BBox) -> CompoundGlyph:
    ''' Read compound glyph from the fontfile. Assumes filepointer is set '''
    fontfile = font.fontfile

    ARG_WORDS = 0x0001
    ARG_XY = 0x0002
    # ROUND = 0x0004
    SCALE = 0x0008
    MORE = 0x0020
    XYSCALE = 0x0040
    TWOBYTWO = 0x0080
    # INSTRUCTIONS = 0x0100
    # METRICS = 0x0200
    # OVERLAP = 0x0400

    glyphidxs = []
    xforms = []
    moreglyphs = True
    while moreglyphs:
        flag = fontfile.readuint16()
        moreglyphs = (flag & MORE) == MORE
        subindex = fontfile.readuint16()
        match = False
        if (flag & ARG_WORDS):
            if (flag & ARG_XY):
                e = fontfile.readint16()
                f = fontfile.readint16()
            else:
                match = True
                e = fontfile.readuint16()
                f = fontfile.readuint16()
        else:
            if (flag & ARG_XY):
                e = fontfile.readint8()
                f = fontfile.readint8()
            else:
                match = True
                e = fontfile.readint8()
                f = fontfile.readint8()

        # Read Transformation
        if (flag & SCALE):
            a = d = fontfile.readshort()
            b = c = 0.
        elif (flag & XYSCALE):
            a = fontfile.readshort()
            d = fontfile.readshort()
            b = c = 0.
        elif (flag & TWOBYTWO):
            a = fontfile.readshort()
            b = fontfile.readshort()
            c = fontfile.readshort()
            d = fontfile.readshort()
        else:
            a = d = 1.
            b = c = 0.

        xforms.append(Xform(a, b, c, d, e, f, match))
        glyphidxs.append(subindex)

    glyphs = []
    for idx in glyphidxs:
        glyphs.append(read_glyph_glyf(idx, font))

    comp = GlyphComp(glyphs, xforms, charbox)
    return CompoundGlyph(index, comp, font)



''' Search system font paths for a font by name '''
from __future__ import annotations
from typing import Sequence, Optional
import os
import sys
import struct
from pathlib import Path

from .fonttypes import FontNames


FONTPATHS: list[Path]
_HOME = Path.home()

if sys.platform == 'win32':
    _WIN = Path(os.environ['WINDIR'])
    FONTPATHS = [
        _WIN.joinpath('Fonts')
    ]
else:
    FONTPATHS = [
        Path('/usr/share/fonts/'),
        Path('/usr/local/share/fonts/'),
        _HOME.joinpath('.local/share/fonts'),
        _HOME.joinpath('.fonts'),
    ]
    if sys.platform == 'darwin':
        FONTPATHS = [
            Path('/Library/Fonts/'),
            Path('/Network/Library/Fonts/'),
            Path('/System/Library/Fonts/'),
            Path('/opt/local/share/fonts'),
            _HOME.joinpath('Library/Fonts'),
        ] + FONTPATHS


def font_list(paths: Optional[Sequence[Path | str]] = None) -> list[Path]:
    ''' List all TTF and OTF files in system font paths '''
    if paths is None:
        searchpaths = FONTPATHS
    else:
        searchpaths = [Path(p) for p in paths] + FONTPATHS

    fontfiles = []
    for fontpath in searchpaths:
        for path, _, fnames in os.walk(fontpath):
            for fname in fnames:
                if fname.lower().endswith('.ttf') or fname.lower().endswith('.otf'):
                    fontfiles.append(Path(path, fname))
    return fontfiles


def readnames(path: Path) -> FontNames:
    ''' Quickly read NAME table from TTF without loading entire font '''
    with open(path, 'rb') as f:
        nameids = [''] * 15  # Empty strings for nameId table

        f.read(4)  # Skip scalartype
        try:
            numtables = struct.unpack('>H', f.read(2))[0]
        except struct.error:
            # empty/invalid font file
            return FontNames(*nameids)

        f.read(6)  # Skip searchrange, entrysel, rangeshift
        tableoffset = -1
        for i in range(numtables):
            try:
                tag = f.read(4).decode()
            except UnicodeDecodeError:
                return FontNames(*nameids)  # invalid font file
            else:
                if tag == 'name':
                    chksum = struct.unpack('>I', f.read(4))[0]
                    tableoffset = struct.unpack('>I', f.read(4))[0]
                    length = struct.unpack('>I', f.read(4))[0]
                    break
                else:
                    f.read(12) # Skip
        if tableoffset == -1:
            return FontNames(*nameids)  # invalid font file

        f.seek(tableoffset)
        namefmt = struct.unpack('>H', f.read(2))[0]
        count = struct.unpack('>H', f.read(2))[0]
        strofst = struct.unpack('>H', f.read(2))[0]

        namerecords = []
        for i in range(count):
            platformId = struct.unpack('>H', f.read(2))[0]
            platformSpecificId = struct.unpack('>H', f.read(2))[0]
            languageId = struct.unpack('>H', f.read(2))[0]
            nameId = struct.unpack('>H', f.read(2))[0]
            length = struct.unpack('>H', f.read(2))[0]
            offset = struct.unpack('>H', f.read(2))[0]
            namerecords.append((platformId, platformSpecificId, languageId,
                                nameId, length, offset))
        for record in namerecords:
            f.seek(tableoffset + strofst + record[5])
            name = f.read(record[4])
            if record[3] < 16:
                if record[0] in [0, 3]:  # Microsoft and Unicode formats
                    nameids[record[3]] = name.decode('utf-16be')

    return FontNames(*nameids)


def system_fonts(paths: Optional[Sequence[str | Path]] = None) -> dict[str | Path, Path]:
    ''' Get dictionary of fontname: fontpath for all system TTF/OTF fonts

        Args:
            paths: Additional paths to search
    '''
    fontdict = {}
    fontlist = font_list(paths)
    for f in fontlist:
        names = readnames(f)
        fontdict[names.name] = f

        if names.subfamily.lower() in ['regular', 'book', 'roman'] and names.family != names.name:
            fontdict[names.family] = f
        fontdict[f'{names.family}, {names.subfamily}'] = f
    return fontdict


def find_font(name: str | Path, paths: Optional[Sequence[str | Path]] = None) -> Optional[Path]:
    ''' Find a font's filename given the name defined in font file.
        Searches all TTF/OTF fonts in (os-dependent) system font paths,
        and the additional paths specified by `paths` parameter.

        Returns None if the font was not found.

        Args:
            name: Name of font, such as "DejaVu Sans"
            paths: Additional paths to search
    '''
    return system_fonts(paths).get(name, None)



''' CFF/Type 2 Charstring Glyphs '''

from __future__ import annotations
from typing import Union, TYPE_CHECKING

import struct
from enum import Enum
import warnings

from .glyph import SimpleGlyph
from .fonttypes import BBox
from .svgpath import Point, Moveto, Lineto, Cubic, SVGOpType

if TYPE_CHECKING:
    from .font import Font


class Operator(Enum):
    ''' Charstring Operators '''
    HSTEM = 1
    # reserved = 2
    VSTEM = 3
    VMOVETO = 4
    RLINETO = 5
    HLINETO = 6
    VLINETO = 7
    RRCURVETO = 8
    # reserved = 9
    CALLSUBR = 10
    RETURN = 11
    # escape = 12
    # reserved = 13
    ENDCHAR = 14
    # reserved = 15, 16, 17
    HSTEMHM = 18
    HINTMASK = 19
    CNTRMASK = 20
    RMOVETO = 21
    HMOVETO = 22
    VSTEMHM = 23
    RCURVELINE = 24
    RLINECURVE = 25
    VVCURVETO = 26
    HHCURVETO = 27
    # reserved = 28/short int
    CALLGSUBR = 29
    VHCURVETO = 30
    HVCURVETO = 31
    # 2-byte codes
    AND = 0x0c03
    OR = 0x0c04
    NOT = 0x0c05
    ABS = 0x0c09
    ADD = 0x0c0a
    SUB = 0x0c0c
    NEG = 0x0c0e
    EQ = 0x0c0f
    DROP = 0x0c12
    PUT = 0x0c14
    GET = 0x0c15
    IFELSE = 0x0c16
    RANDOM = 0x0c17
    MUL = 0x0c18
    SQRT = 0x0c1a
    DUP = 0x0c1b
    EXCH = 0x0c1c
    INDEX = 0x0c1d
    ROLL = 0x0c1e
    HFLEX = 0x0c22
    FLEX = 0x0c23
    HFLEX1 = 0x0c24
    FLEX1 = 0x0c25
    WIDTH = -999  # Not numbered in CFF spec


def readreal(buf: bytes) -> tuple[float, int]:
    ''' Read real/float value from buffer. Returns (value, #bytes read) '''
    nibbles = {
        0x0a: '.',
        0x0b: 'E',
        0x0c: 'E-',
        # 0x0d: RESERVED,
        0x0e: '-',
        0x0f: ''}  # END

    numstr = ''
    i = 0
    while True:
        nib1 = (buf[i] & 0xF0) >> 4
        nib2 = (buf[i] & 0x0F)
        if nib1 == 0x0f:
            break
        numstr += nibbles.get(nib1, str(nib1))
        if nib2 == 0x0f:
            break
        numstr += nibbles.get(nib2, str(nib2))
        i += 1

    return float(numstr), i+1


def read_glyph_cff(glyphid: int, font: Font) -> SimpleGlyph:
    ''' Read a glyph from the CFF table. '''
    charstr = font.cffdata.charstr_index[glyphid]  # type: ignore
    ops, width, bbox = charstr2path(charstr, font.cffdata)  # type: ignore
    glyph = SimpleGlyph(glyphid, ops, bbox, font)
    return glyph


def charstr2path(charstr: bytes, cff: CFF) -> tuple[list[SVGOpType], float, BBox]:
    ''' Convert the charstring operators into SVG path elements '''
    operators: list[SVGOpType] = []
    width = cff.defaultwidth
    p = Point(0, 0)
    chstate = CharString(charstr, cff)
    op = None
    for op, value in zip(chstate.operators, chstate.operands):
        if op == Operator.WIDTH:
            width = cff.nominalwidth + int(value[0])
        elif op == Operator.RMOVETO:
            p = p + Point(value[0], value[1])
            operators.append(Moveto(p))
        elif op == Operator.HMOVETO:
            p = p + Point(value[0], 0)
            operators.append(Moveto(p))
        elif op == Operator.VMOVETO:
            p = p + Point(0, value[0])
            operators.append(Moveto(p))
        elif op == Operator.HLINETO:
            for i, val in enumerate(value):
                if (i % 2) == 0:  # Alternate Horizontal and Vertical lines
                    p = p + Point(val, 0)
                else:
                    p = p + Point(0, val)
                operators.append(Lineto(p))

        elif op == Operator.VLINETO:
            for i, val in enumerate(value):
                if (i % 2) == 0:  # Alternate Horizontal and Vertical lines
                    p = p + Point(0, val)
                else:
                    p = p + Point(val, 0)
                operators.append(Lineto(p))

        elif op == Operator.RLINETO:
            for dx, dy in zip(value[::2], value[1::2]):
                p = p + Point(dx, dy)
                operators.append(Lineto(p))

        elif op in [Operator.RRCURVETO, Operator.RCURVELINE]:
            while len(value) > 2:
                dxa, dya, dxb, dyb, dxc, dyc, *_ = value
                p1 = p + Point(dxa, dya)
                p2 = p1 + Point(dxb, dyb)
                p = p2 + Point(dxc, dyc)
                operators.append(Cubic(p1, p2, p))
                value = value[6:]
            if op == Operator.RCURVELINE:
                assert len(value) >= 2
                # RCURVELINE ends with a line
                p = p + Point(value[0], value[1])
                operators.append(Lineto(p))

        elif op == Operator.RLINECURVE:
            while len(value) > 6:
                # Lines until last 6 arguments
                p = p + Point(value[0], value[1])
                operators.append(Lineto(p))
                value = value[2:]
            # Last 6 define Bezier
            dxa, dya, dxb, dyb, dxc, dyc, *_ = value
            p1 = p + Point(dxa, dya)
            p2 = p1 + Point(dxb, dyb)
            p = p2 + Point(dxc, dyc)
            operators.append(Cubic(p1, p2, p))

        elif op == Operator.HHCURVETO:
            if len(value) % 4 >= 1:
                dya = value[0]
                value = value[1:]
            else:
                dya = 0
            while len(value) >= 4:
                dxa, dxb, dyb, dxc, *_ = value
                p1 = p + Point(dxa, dya)
                p2 = p1 + Point(dxb, dyb)
                p = p2 + Point(dxc, 0)
                operators.append(Cubic(p1, p2, p))
                dya = 0
                value = value[4:]

        elif op == Operator.VVCURVETO:
            if len(value) % 4 >= 1:
                dxa = value[0]
                value = value[1:]
            else:
                dxa = 0
            while len(value) >= 4:
                dya, dxb, dyb, dyc, *_ = value
                p1 = p + Point(dxa, dya)
                p2 = p1 + Point(dxb, dyb)
                p = p2 + Point(0, dyc)
                operators.append(Cubic(p1, p2, p))
                value = value[4:]
                dxa = 0

        elif op == Operator.HVCURVETO:
            if len(value) % 8 >= 4:
                dx1, dx2, dy2, dy3, *_ = value
                p1 = p + Point(dx1, 0)
                p2 = p1 + Point(dx2, dy2)
                p = p2 + Point(0, dy3)
                if len(value) == 5:
                    p = p + Point(value[-1], 0)
                operators.append(Cubic(p1, p2, p))
                value = value[4:]
                while len(value) >= 8:
                    dya, dxb, dyb, dxc, dxd, dxe, dye, dyf, *_ = value
                    if len(value) == 9:
                        dxf = value[-1]
                    else:
                        dxf = 0

                    p1 = p + Point(0, dya)
                    p2 = p1 + Point(dxb, dyb)
                    p = p2 + Point(dxc, 0)
                    operators.append(Cubic(p1, p2, p))
                    p1 = p + Point(dxd, 0)
                    p2 = p1 + Point(dxe, dye)
                    p = p2 + Point(dxf, dyf)
                    operators.append(Cubic(p1, p2, p))
                    value = value[8:]

            else:
                while len(value) >= 8:
                    dxa, dxb, dyb, dyc, dyd, dxe, dye, dxf, *_ = value
                    if len(value) == 9:
                        dyf = value[-1]
                    else:
                        dyf = 0

                    p1 = p + Point(dxa, 0)
                    p2 = p1 + Point(dxb, dyb)
                    p = p2 + Point(0, dyc)
                    operators.append(Cubic(p1, p2, p))
                    p1 = p + Point(0, dyd)
                    p2 = p1 + Point(dxe, dye)
                    p = p2 + Point(dxf, dyf)
                    operators.append(Cubic(p1, p2, p))
                    value = value[8:]

        elif op == Operator.VHCURVETO:
            if len(value) % 8 >= 4:
                dy1, dx2, dy2, dx3, *_ = value
                p1 = p + Point(0, dy1)
                p2 = p1 + Point(dx2, dy2)
                p = p2 + Point(dx3, 0)
                if len(value) == 5:
                    p = p + Point(0, value[-1])
                operators.append(Cubic(p1, p2, p))
                value = value[4:]
                while len(value) >= 8:
                    dxa, dxb, dyb, dyc, dyd, dxe, dye, dxf, *_ = value
                    if len(value) == 9:
                        dyf = value[-1]
                    else:
                        dyf = 0

                    p1 = p + Point(dxa, 0)
                    p2 = p1 + Point(dxb, dyb)
                    p = p2 + Point(0, dyc)
                    operators.append(Cubic(p1, p2, p))
                    p1 = p + Point(0, dyd)
                    p2 = p1 + Point(dxe, dye)
                    p = p2 + Point(dxf, dyf)
                    operators.append(Cubic(p1, p2, p))
                    value = value[8:]

            else:
                while len(value) >= 8:
                    dya, dxb, dyb, dxc, dxd, dxe, dye, dyf, *_ = value
                    if len(value) == 9:
                        dxf = value[-1]
                    else:
                        dxf = 0

                    p1 = p + Point(0, dya)
                    p2 = p1 + Point(dxb, dyb)
                    p = p2 + Point(dxc, 0)
                    operators.append(Cubic(p1, p2, p))
                    p1 = p + Point(dxd, 0)
                    p2 = p1 + Point(dxe, dye)
                    p = p2 + Point(dxf, dyf)
                    operators.append(Cubic(p1, p2, p))
                    value = value[8:]

        elif op == Operator.FLEX:
            dx1, dy1, dx2, dy2, dx3, dy3, dx4, dy4, dx5, dy5, dx6, dy6, *_ = value
            p1 = p + Point(dx1, dy1)
            p2 = p1 + Point(dx2, dy2)
            p3 = p2 + Point(dx3, dy3)
            p4 = p3 + Point(dx4, dy4)
            p5 = p4 + Point(dx5, dy5)
            p = p5 + Point(dx6, dy6)
            operators.append(Cubic(p1, p2, p3))
            operators.append(Cubic(p4, p5, p))

        elif op == Operator.FLEX1:
            dx1, dy1, dx2, dy2, dx3, dy3, dx4, dy4, dx5, dy5, d6, *_ = value
            dx = dx1+dx2+dx3+dx4+dx5
            dy = dy1+dy2+dy3+dy4+dy5

            p1 = p + Point(dx1, dy1)
            p2 = p1 + Point(dx2, dy2)
            p3 = p2 + Point(dx3, dy3)
            p4 = p3 + Point(dx4, dy4)
            p5 = p4 + Point(dx5, dy5)
            if abs(dx) > abs(dy):
                p = p5 + Point(d6, 0)
            else:
                p = p5 + Point(0, d6)
            operators.append(Cubic(p1, p2, p3))
            operators.append(Cubic(p4, p5, p))

        elif op == Operator.HFLEX:
            dx1, dx2, dy2, dx3, dx4, dx5, dx6, *_ = value
            p1 = p + Point(dx1, 0)
            p2 = p1 + Point(dx2, dy2)
            p3 = p2 + Point(dx3, 0)
            p4 = p3 + Point(dx4, 0)
            p5 = p4 + Point(dx5, 0)
            p = p5 + Point(dx6, 0)
            operators.append(Cubic(p1, p2, p3))
            operators.append(Cubic(p4, p5, p))

        elif op == Operator.HFLEX1:
            dx1, dy1, dx2, dy2, dx3, dx4, dx5, dy5, dx6, *_ = value
            p1 = p + Point(dx1, dy1)
            p2 = p1 + Point(dx2, dy2)
            p3 = p2 + Point(dx3, 0)
            p4 = p3 + Point(dx4, 0)
            p5 = p4 + Point(dx5, dy5)
            p = p5 + Point(dx6, 0)
            operators.append(Cubic(p1, p2, p3))
            operators.append(Cubic(p4, p5, p))

        elif op not in [Operator.HSTEM, Operator.HINTMASK, Operator.VSTEM,
                        Operator.VSTEMHM, Operator.HSTEMHM, Operator.RETURN,
                        Operator.CNTRMASK, Operator.ENDCHAR]:
            raise NotImplementedError(f'Operator {op} not implemented')

    if op != Operator.ENDCHAR:
        warnings.warn('Glyph has no ENDCHAR')

    try:
        xmin = min(op.xmin() for op in operators)
        xmax = max(op.xmax() for op in operators)
        ymin = min(op.ymin() for op in operators)
        ymax = max(op.ymax() for op in operators)
    except ValueError:  # no operators
        ymin = ymax = xmin = xmax = 0
    return operators, width, BBox(xmin, xmax, ymin, ymax)


class CharString:
    ''' CharString reader '''
    def __init__(self, buf: bytes, cff: CFF):
        self.buf = buf
        self.cff = cff

        self.stack: list[float] = []
        self.operators: list[Union[Operator, int]] = []
        self.operands: list[list[float]] = []
        self.nhints = 0
        self.readcharstr(buf)

    def append(self, operator: Operator):
        ''' Append an operator, and clear the stack '''
        self.operators.append(operator)
        self.operands.append(self.stack)
        self.stack = []

    def addwidth(self):
        ''' Add width operator '''
        self.operators.append(Operator.WIDTH)
        self.operands.append([self.stack[0]])
        self.stack = self.stack[1:]

    @property
    def lenstack(self):
        ''' Length of stack '''
        return len(self.stack)

    def readcharstr(self, buf: bytes):
        ''' Read charstring bytes into operators/operands lists '''
        key: Union[Operator, int]
        while len(buf) > 0:
            if buf[0] <= 27 or 29 <= buf[0] <= 31:  # Operator
                key = buf[0]
                nbytes = 1
                if buf[0] == 12:
                    key = struct.unpack_from('>H', buf)[0]
                    nbytes = 2

                try:
                    key = Operator(key)
                except ValueError as exc:
                    break
                    # raise NotImplementedError(f'Unimplemented KEY {key}') from exc

                if len(self.operators) == 0:
                    # First operator can have extra width parameter.
                    # have to deduce its presence based on the
                    # operator and number of values in stack
                    if key in [Operator.CNTRMASK, Operator.ENDCHAR] and self.lenstack == 1:
                        self.addwidth()
                    elif ((key in [Operator.HMOVETO, Operator.VMOVETO] and self.lenstack > 1) or
                          (key in [Operator.RMOVETO] and self.lenstack > 2)):
                        self.addwidth()
                    elif (key in [Operator.HSTEM, Operator.HSTEMHM, Operator.VSTEM,
                                  Operator.VSTEMHM, Operator.HINTMASK] and
                          self.lenstack % 2 != 0):
                        self.addwidth()

                if key in [Operator.CALLSUBR, Operator.CALLGSUBR]:
                    if key == Operator.CALLSUBR:
                        sublist = self.cff.localsubs
                    else:
                        sublist = self.cff.globalsub

                    if self.cff.topdict.get('charstringtype', 2) == 1:
                        bias = 0
                    elif len(sublist) < 1240:
                        bias = 107
                    elif len(sublist) < 33900:
                        bias = 1131
                    else:
                        bias = 32768
                    idx = int(self.stack[-1] + bias)

                    subchstr = sublist[idx]
                    self.stack = self.stack[:-1]  # remove sub# from stack
                    self.readcharstr(subchstr)
                    buf = buf[nbytes:]
                    continue

                elif key in [Operator.HINTMASK, Operator.CNTRMASK]:
                    if (len(self.operators) > 0
                            and self.operators[-1] == Operator.HSTEMHM
                            and self.lenstack > 0):
                        # Implied VSTEM operator
                        self.nhints += self.lenstack // 2
                        self.append(Operator.VSTEMHM)
                    elif (len(self.operators) == 0
                            or (len(self.operators) == 1 and self.operators[0] == Operator.WIDTH)):
                        self.nhints += self.lenstack // 2
                        self.append(Operator.VSTEMHM)
                    # N-bits for the N hint masks just read in
                    hintbytes = self.nhints + 7 >> 3
                    self.stack = list(buf[1:hintbytes+1])
                    nbytes = hintbytes+1

                elif key in [Operator.HSTEM, Operator.HSTEMHM,
                             Operator.VSTEM, Operator.VSTEMHM]:
                    self.nhints += self.lenstack // 2

                if key not in [Operator.RETURN]:
                    self.append(key)

            # Opearands (numeric values)
            elif buf[0] == 28:
                self.stack.append(struct.unpack_from('>h', buf[1:])[0])  # signed short
                nbytes = 3
            elif 32 <= buf[0] <= 246:
                self.stack.append(buf[0] - 139)
                nbytes = 1
            elif 247 <= buf[0] <= 250:
                self.stack.append((buf[0]-247)*256 + buf[1] + 108)
                nbytes = 2
            elif 251 <= buf[0] <= 254:
                self.stack.append(-(buf[0]-251)*256 - buf[1] - 108)
                nbytes = 2
            elif buf[0] == 255:
                self.stack.append(struct.unpack_from('>l', buf[1:])[0] / 0x10000)
                nbytes = 5
            else:
                raise ValueError('Bad encoding byte: ' + str(buf[0]))

            buf = buf[nbytes:]


def readdict(buf: bytes) -> dict:
    ''' Read a CFF dictionary structure from the buffer '''
    data: dict[Union[int, str], Union[None, list[float], float]] = {}
    value: list[float] = []
    key: Union[int, str]
    while len(buf) > 0:
        # Operator (key)
        if buf[0] <= 21:
            key = buf[0]
            i = 1
            if buf[0] == 12:
                key = struct.unpack_from('>h', buf)[0]
                i = 2
            if len(value) == 0:
                data[key] = None
            elif len(value) == 1:
                data[key] = value[0]
            else:
                data[key] = value
            value = []

        # Opearand (value)
        elif buf[0] == 28:
            value.append(struct.unpack_from('>h', buf[1:])[0])  # signed short
            i = 3
        elif buf[0] == 29:
            value.append(struct.unpack_from('>l', buf[1:])[0])  # signed long
            i = 5
        elif 32 <= buf[0] <= 246:
            value.append(buf[0] - 139)
            i = 1
        elif 247 <= buf[0] <= 250:
            value.append((buf[0]-247)*256 + buf[1] + 108)
            i = 2
        elif 251 <= buf[0] <= 254:
            value.append(-(buf[0]-251)*256 - buf[1] - 108)
            i = 2
        elif buf[0] == 30:
            v, nbytes = readreal(buf[1:])
            value.append(v)
            i = (nbytes+1)
        else:
            warnings.warn('Bad encoding byte: ' + str(buf[0]))
            value.append(0)
            i = 1

        buf = buf[i:]
    return data


class CFF:
    ''' Compact Font Format table info '''
    def __init__(self, font: Font):
        self.cffofst = font.tables['CFF '].offset
        self.font = font
        self.fontfile = self.font.fontfile

        self.major = self.fontfile.readuint8(self.cffofst)
        self.minor = self.fontfile.readuint8()
        self.headsize = self.fontfile.readuint8()
        self.offsize = self.fontfile.readuint8()
        self.names = self.readindex()
        topdict_bytes = self.readindex()
        self.strings = self.readindex()
        self.globalsub: list[bytes] = self.readindex()
        self.topdicts: list[dict] = [self.read_topdict(readdict(t)) for t in topdict_bytes]
        self.set_topdict(0)

    def set_topdict(self, idx: int = 0) -> None:
        ''' Set which TOP DICT to use and load the corresponding
            PRIVATE DICT values.
        '''
        topdict = self.topdicts[idx]
        self.topdict = topdict
        charstr_ofst = self.cffofst + topdict['charstrings']
        self.charstr_index: list[bytes] = self.readindex(charstr_ofst)
        self.nglyphs = len(self.charstr_index)

        self.font.fontfile.seek(self.cffofst + topdict['private'][1])
        pdictbytes = self.font.fontfile.read(topdict['private'][0])
        self.privatedict = self.read_topdict(readdict(pdictbytes))
        self.localsubs: list[bytes] = []
        if 'subrs' in self.privatedict:
            self.localsubs = self.readindex(
                self.cffofst + topdict['private'][1] + self.privatedict['subrs'])

    @property
    def defaultwidth(self) -> int:
        ''' Get default glyph width '''
        return self.privatedict.get('defaultwidthx', 0)

    @property
    def nominalwidth(self) -> int:
        ''' Get nominal glyph width '''
        return self.privatedict.get('nominalwidthx', 0)

    def getstring(self, sid: int) -> bytes:
        ''' Get a string by String ID '''
        nstandardstrings = 391
        if sid > nstandardstrings:
            return self.strings[sid-nstandardstrings]
        else:
            return b'standard string'  # TODO

    def readindex(self, offset=None) -> list[bytes]:
        ''' Read an INDEX structure from the CFF data '''
        count = self.fontfile.readuint16(offset)
        offsize = self.fontfile.readuint8()  # should be 1-4
        if offsize == 0:
            return []

        offsets = []
        for i in range(count+1):
            if offsize == 1:
                offsets.append(self.fontfile.readuint8())
            elif offsize == 2:
                offsets.append(self.fontfile.readuint16())
            elif offsize == 3:
                offsets.append(self.fontfile.readuint24())
            elif offsize == 4:
                offsets.append(self.fontfile.readuint32())
            else:
                raise ValueError('Incorrect offset size ' + str(offsize))

        dataofst = self.fontfile.tell()-1  # Start of Object data (minus 1)
        values = []
        for i in range(count):
            self.fontfile.seek(dataofst + offsets[i])
            values.append(self.fontfile.read(offsets[i+1]-offsets[i]))
        return values

    def read_topdict(self, d: dict) -> dict:
        ''' Translate raw CFF dictionary into topdict with meaningful keys '''
        topdict_entries = {
            0: ('version', 'sid'),
            1: ('notice', 'sid'),
            2: ('fullname', 'sid'),
            3: ('familyname', 'sid'),
            4: ('weight', 'number'),
            5: ('fontbbox', 'array'),
            13: ('uniqueid', 'number'),
            14: ('xuid', 'array'),
            15: ('charset', 'number'),
            16: ('encoding', 'number'),
            17: ('charstrings', 'number'),
            18: ('private', 'array'),
            0x0c00: ('copyright', 'sid'),
            0x0c01: ('isfixedpitch', 'number'),
            0x0c02: ('italicangle', 'number'),
            0x0c03: ('underlineposition', 'number'),
            0x0c04: ('underlinethickness', 'number'),
            0x0c05: ('painttype', 'number'),
            0x0c06: ('charstringtype', 'number'),
            0x0c07: ('fontmatrix', 'array'),
            0x0c08: ('strokewidth', 'number'),
            0x0c14: ('syntheticbase', 'number'),
            0x0c15: ('postscript', 'sid'),
            0x0c16: ('basefontname', 'sid'),
            0x0c17: ('basefontblend' 'array'),
            # Private Dict Keys
            6: ('bluevalues', 'array'),
            7: ('otherblues', 'array'),
            8: ('familyblues', 'array'),
            9: ('familyotherblues', 'array'),
            10: ('stdhw', 'number'),
            11: ('stdvw', 'number'),
            19: ('subrs', 'number'),
            20: ('defaultwidthx', 'number'),
            21: ('nominalwidthx', 'number'),
            0x0c09: ('bluescale', 'number'),
            0x0c0a: ('blueshift', 'number'),
            0x0c0b: ('bluefuzz', 'number'),
            0x0c0c: ('stemsnaph', 'array'),
            0x0c0d: ('stemsnapv', 'array'),
            0x0c0e: ('forcebold', 'number'),
            0x0c11: ('languagegroup', 'number'),
            0x0c12: ('expansionfactor', 'number'),
            0x0c13: ('initialrandomseed', 'number'),
            # CIDfont
            0x0c1e: ('ROS', 'array'),
            0x0c1f: ('CIDFontVersion', 'number'),
            0x0c20: ('CIDFontRevision', 'number'),
            0x0c21: ('CIDFontType', 'number'),
            0x0c22: ('CIDCount', 'number'),
            0x0c23: ('UIDBase', 'number'),
            0x0c24: ('FDArray', 'array'),
            0x0c25: ('FDSelect', 'number'),
            0x0c26: ('FontName', 'sid')}

        newd = {}
        for key, value in d.items():
            newkey, dtype = topdict_entries.get(key, (key, 'number'))
            if newkey:
                if dtype == 'sid':
                    value = self.getstring(value)
                newd[newkey] = value

        assert newd.get('charstringtype', 2) == 2
        if 'ROS' in newd:
            raise NotImplementedError('CID fonts not implemented.')
        return newd



''' Glyph Positioning System (GPOS) tables '''

from __future__ import annotations
from typing import Optional, Union
import logging
from collections import namedtuple

from .fontread import FontReader
from .tables import Coverage, ClassDef, Feature, Script, Language
from .glyph import SimpleGlyph


PositionDelta = namedtuple('PositionDelta', ['dx', 'dy', 'dadvance'])

# Features that cant' be turned off
PERM_FEATURES = ['mark', 'mkmk', 'curs']

# Features that are on by default
ON_FEATURES = ['kern']


class Gpos:
    ''' Glyph Positioning System Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        self.language = Language()
        self.fontfile.seek(self.ofst)

        self.vermajor = self.fontfile.readuint16()
        self.verminor = self.fontfile.readuint16()
        scriptofst = self.fontfile.readuint16()
        featureofst = self.fontfile.readuint16()
        lookupofst = self.fontfile.readuint16()
        if self.verminor > 0:
            self.variationofst = self.fontfile.readuint32()
            logging.warning('GPOS has feature variations - unimplemented')

        # Read scripts
        self.scripts = {}
        if scriptofst != 0:
            scriptlisttableloc = self.ofst + scriptofst
            scriptcnt = self.fontfile.readuint16(scriptlisttableloc)
            for i in range(scriptcnt):
                tag = self.fontfile.read(4).decode()
                self.scripts[tag] = Script(
                    tag,
                    self.fontfile.readuint16() + scriptlisttableloc,
                    self.fontfile)

        # Read features
        featurelist = []
        if featureofst != 0:
            featurelisttableloc = self.ofst + featureofst
            featurecnt = self.fontfile.readuint16(featurelisttableloc)
            for i in range(featurecnt):
                featurelist.append(Feature(
                    self.fontfile.read(4).decode(),
                    self.fontfile.readuint16() + featurelisttableloc,
                    self.fontfile))

        # Read Lookups
        self.lookups = []
        if lookupofst != 0:
            lookuplisttableloc = self.ofst + lookupofst
            lookupcnt = self.fontfile.readuint16(lookuplisttableloc)
            for i in range(lookupcnt):
                self.lookups.append(GposLookup(
                    self.fontfile.readuint16() + lookuplisttableloc,
                    self.fontfile))

        # Put everything in a dictionary for access
        self.features = {}
        for scrname, script in self.scripts.items():
            langdict = {}
            for langname, lang in script.languages.items():
                langfeatures = [featurelist[i] for i in lang.featureidxs]
                featnames = [f.tag for f in langfeatures]
                featdict = {}
                for feat in featnames:
                    lookups = langfeatures[featnames.index(feat)].lookupids
                    tables = [self.lookups[i] for i in lookups]
                    featdict[feat] = tables
                langdict[langname] = featdict
            self.features[scrname] = langdict

        if self.features and 'latn' not in self.features:
            self.language.script = list(self.features.keys())[0]

    def features_available(self):
        ''' Dictionary of features active in the current script/language system '''
        return self.features.get(self.language.script, {}).get(self.language.language, {})

    def init_user_features(self) -> dict[str, bool]:
        ''' Initialize features that can be set by user '''
        avail = list(self.features_available().keys())
        avail = [feat for feat in avail if feat not in PERM_FEATURES]
        return {feat: feat in ON_FEATURES for feat in avail}

    def position(self, glyphs: list[SimpleGlyph], features: dict[str, bool]) -> list[tuple[int, int]]:
        ''' Calculate x, y position of each glyph using the given features '''
        gids = [glyph.index for glyph in glyphs]
        delta: list[PositionDelta] = [PositionDelta(0,0,0) for _ in gids]
        advances = [glyph.advance() for glyph in glyphs]

        feattable = self.features_available()
        for feat in feattable.keys():
            if feat in PERM_FEATURES or features.get(feat, False):        
                tables = feattable.get(feat, [])
                for table in tables:
                    ddelta = table.adjust(*gids, advances=advances, lookups=self.lookups)
                    delta = merge_deltas(delta, ddelta)

        xy: list[tuple[int, int]] = []
        x = 0
        for d, glyph in zip(delta, glyphs):
            x += d.dx
            xy.append((x, d.dy))
            x += glyph.advance() + d.dadvance
        return xy

    def __repr__(self):
        return f'<GPOS Table v{self.vermajor}.{self.verminor}>'


class GposLookup:
    ''' GPOS Lookup Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        RIGHT_TO_LEFT = 0x0001
        IGNORE_BASE_GLYPHS = 0x0002
        IGNORE_LIGATURES = 0x0004
        IGNORE_MARKS = 0x0008
        USE_MARK_FILTERING_SET = 0x0010
        MARK_ATTACHMENT_TYPE_MASK = 0xFF00

        self.fontfile.seek(self.ofst)
        self.type = self.fontfile.readuint16()
        self.flag = self.fontfile.readuint16()
        subtablecnt = self.fontfile.readuint16()
        self.tableofsts = []
        for _ in range(subtablecnt):
            self.tableofsts.append(self.fontfile.readuint16())
        self.markfilterset = None
        if self.flag & USE_MARK_FILTERING_SET:
            self.markfilterset = self.fontfile.readuint16()

        self.subtables: list[GposSubtable] = []
        for tblofst in self.tableofsts:
            tabletype = self.type
            fmt = self.fontfile.readuint16(self.ofst + tblofst)
            if self.type == 9:  # Extension table - converts to another type
                assert fmt == 1
                tabletype = self.fontfile.readuint16()
                tblofst += self.fontfile.readuint32()
                fmt = self.fontfile.readuint16(self.ofst+tblofst)

            if tabletype == 1 and fmt == 1:
                self.subtables.append(SingleAdjustmentSubtable(
                    tblofst + self.ofst, self.fontfile))
            elif tabletype == 1 and fmt == 2:
                self.subtables.append(SingleAdjustmentSubtable2(
                    tblofst + self.ofst, self.fontfile))
            elif tabletype == 2:  # Pair adjustment positioning
                self.subtables.append(PairAdjustmentSubtable(
                        tblofst + self.ofst, self.fontfile))
            elif tabletype == 4:  # Mark-to-base
                self.subtables.append(MarkToBaseSubtable(
                        tblofst + self.ofst, self.fontfile))
            elif tabletype == 6:  # Mark-to-mark
                self.subtables.append(MarkToMarkSubtable(
                        tblofst + self.ofst, self.fontfile))
            else:
                logging.debug('Unimplemented GPOS Lookup Type %s', self.type)

        self.fontfile.seek(fileptr)  # Put file pointer back

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        deltas: list[PositionDelta] = [PositionDelta(0,0,0) for _ in glyphids]
        for subtable in self.subtables:
            d = subtable.adjust(*glyphids, advances=advances, lookups=lookups)
            deltas = merge_deltas(deltas, d)
        return deltas

    def __repr__(self):
        return f'<GPOSLookup Type {self.type} {hex(self.ofst)}>'


def merge_delta(d1: PositionDelta, d2: PositionDelta) -> PositionDelta:
    return PositionDelta(d1.dx+d2.dx,
                         d1.dy+d2.dy,
                         d1.dadvance+d2.dadvance)


def merge_deltas(deltas1: list[PositionDelta], deltas2: list[PositionDelta]) -> list[PositionDelta]:
    return [merge_delta(d1, d2)
            for d1, d2 in zip(deltas1, deltas2)]


def valuerec_to_delta(vrec: dict[str, int]|None) -> PositionDelta:
    if vrec is None:
        return PositionDelta(0, 0, 0)
    return PositionDelta(
        vrec.get('x', 0),
        vrec.get('y', 0),
        vrec.get('xadvance', 0))


class GposSubtable:
    ''' Base class for GPOS subtables (mostly for type hints) '''
    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        return []


class SingleAdjustmentSubtable(GposSubtable):
    ''' Single adjustment subtable (GPOS Lookup 1.1) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        fmt = self.fontfile.readuint16()
        assert fmt == 1
        self.covofst = self.fontfile.readuint16()
        valueformat = self.fontfile.readuint16()
        self.valuerecord = self.fontfile.readvaluerecord(valueformat)
        self.coverage = Coverage(self.covofst+self.ofst, self.fontfile)
        self.fontfile.seek(fileptr)  # Put file pointer back

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        adjusts: list[PositionDelta] = []

        for gid in glyphids:
            if self.coverage.covidx(gid) is not None:
                pos = PositionDelta(
                    self.valuerecord.get('x', 0),
                    self.valuerecord.get('y', 0),
                    self.valuerecord.get('xadvance', 0))
                logging.debug('Positioning glyph %s: (%s, %s, %s)',
                              gid, pos.dx, pos.dy, pos.dadvance)

            else:
                pos = PositionDelta(0, 0, 0)
            adjusts.append(pos)
        return adjusts


class SingleAdjustmentSubtable2(GposSubtable):
    ''' Single adjustment subtable (GPOS Lookup 1.2) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        fmt = self.fontfile.readuint16()
        assert fmt == 2
        self.covofst = self.fontfile.readuint16()
        valueformat = self.fontfile.readuint16()
        valuecount = self.fontfile.readuint16()
        self.valuerecords = []
        for _ in range(valuecount):
            self.valuerecords.append(self.fontfile.readvaluerecord(valueformat))
        self.coverage = Coverage(self.covofst+self.ofst, self.fontfile)
        self.fontfile.seek(fileptr)  # Put file pointer back

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        adjusts: list[PositionDelta] = []

        for gid in glyphids:
            if (covidx := self.coverage.covidx(gid) is not None):
                pos = PositionDelta(
                    self.valuerecords[covidx].get('x', 0),
                    self.valuerecords[covidx].get('y', 0),
                    self.valuerecords[covidx].get('xadvance', 0))
                logging.debug('Positioning glyph %s: (%s, %s, %s)',
                              gid, pos.dx, pos.dy, pos.dadvance)

            else:
                pos = PositionDelta(0, 0, 0)
            adjusts.append(pos)
        return adjusts


class PairAdjustmentSubtable(GposSubtable):
    ''' Pair Adjustment Table (GPOS Lookup Type 2)
        Informs kerning between pairs of glyphs
    '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.posformat = self.fontfile.readuint16()
        self.covofst = self.fontfile.readuint16()
        self.valueformat1 = self.fontfile.readuint16()
        self.valueformat2 = self.fontfile.readuint16()

        if self.posformat == 1:
            pairsetcount = self.fontfile.readuint16()
            pairsetofsts = []
            for i in range(pairsetcount):
                pairsetofsts.append(self.fontfile.readuint16())

            PairValue = namedtuple('PairValue', ['secondglyph', 'value1', 'value2'])
            
            self.pairsets = []
            for i in range(pairsetcount):
                self.fontfile.seek(pairsetofsts[i] + self.ofst)
                paircnt = self.fontfile.readuint16()
                pairs = []
                for p in range(paircnt):
                    pairs.append(PairValue(
                        self.fontfile.readuint16(),
                        self.fontfile.readvaluerecord(self.valueformat1),
                        self.fontfile.readvaluerecord(self.valueformat2)))
                self.pairsets.append(pairs)

        elif self.posformat == 2:
            classdef1ofst = self.fontfile.readuint16()
            classdef2ofst = self.fontfile.readuint16()
            class1cnt = self.fontfile.readuint16()
            class2cnt = self.fontfile.readuint16()

            self.classrecords = []
            for i in range(class1cnt):
                class2recs = []
                for j in range(class2cnt):
                    class2recs.append(
                        (self.fontfile.readvaluerecord(self.valueformat1),
                         self.fontfile.readvaluerecord(self.valueformat2)))
                self.classrecords.append(class2recs)

            self.class1def = ClassDef(self.ofst + classdef1ofst, self.fontfile)
            self.class2def = ClassDef(self.ofst + classdef2ofst, self.fontfile)

        else:
            raise ValueError('Invalid posformat in PairAdjustmentSubtable')

        self.coverage = Coverage(self.covofst+self.ofst, self.fontfile)
        self.fontfile.seek(fileptr)  # Put file pointer back

    def get_adjust(self, glyph1: int, glyph2: int) -> tuple[Optional[dict], Optional[dict]]:
        ''' Get kerning adjustment for glyph1 and glyph2 pair '''
        v1 = v2 = None

        # Look up first glyph in coverage table
        covidx = self.coverage.covidx(glyph1)
        if covidx is not None:

            # Look up second glyph
            if self.posformat == 1:
                for p in self.pairsets[covidx]:
                    if p.secondglyph == glyph2:
                        v1 = p.value1
                        v2 = p.value2
                        break

            else:
                c1 = self.class1def.get_class(glyph1)
                c2 = self.class2def.get_class(glyph2)
                if c1 is not None and c2 is not None:
                    v1, v2 = self.classrecords[c1][c2]

        return v1, v2

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        deltas: list[PositionDelta] = [PositionDelta(0,0,0)]
        valuerec1 = valuerec2 = None
        for gid, gid0 in zip(glyphids[1:], glyphids[:-1]):
            covidx = self.coverage.covidx(gid0)
            if covidx is None:
                deltas.append(PositionDelta(0, 0, 0))
            else:
                if self.posformat == 1:
                    for p in self.pairsets[covidx]:
                        if p.secondglyph == gid:
                            valuerec1 = p.value1
                            valuerec2 = p.value2
                            break
                else:
                    c1 = self.class1def.get_class(gid0)
                    c2 = self.class2def.get_class(gid)
                    if c1 is not None and c2 is not None:
                        valuerec1, valuerec2 = self.classrecords[c1][c2]

                deltas[-1] = merge_delta(deltas[-1], valuerec_to_delta(valuerec1))
                deltas.append(valuerec_to_delta(valuerec2))

        return deltas

    def __repr__(self):
        return f'<PairAdjustmentSubtable {hex(self.ofst)}>'


def read_markarray_table(ofst, fontfile):
    ''' Read MarkArray table from font file '''
    cnt = fontfile.readuint16(ofst)
    MarkRecord = namedtuple('MarkRecord', ['markclass', 'anchortable'])
    markrecords = []
    for _ in range(cnt):
        markclass = fontfile.readuint16()
        anchorofst = fontfile.readuint16()
        ptr = fontfile.tell()
        anchortable = read_anchortable(ofst+anchorofst, fontfile)
        markrecords.append(MarkRecord(markclass, anchortable))
        ptr = fontfile.seek(ptr)
    return markrecords


def read_anchortable(ofst, fontfile):
    ''' Read anchor table from font file '''
    Anchor = namedtuple('Anchor', ['x', 'y', 'anchorpoint', 'xofst', 'yofst'])
    fmt = fontfile.readuint16(ofst)
    point = None
    xofst = yofst = None
    x = fontfile.readint16()
    y = fontfile.readint16()
    if fmt == 2:
        point = fontfile.readuint16()
    elif fmt == 3:
        xofst = fontfile.readuint16()
        yofst = fontfile.readuint16()
        logging.warning('Anchor references Device Table - unimplemented')
    return Anchor(x, y, point, xofst, yofst)


def read_basearray(ofst, fontfile, markclasscount):
    ''' Read BaseArray table used by Mark-to-Base Lookup '''
    basecnt = fontfile.readuint16(ofst)
    basearray = []
    for i in range(basecnt):
        anchortables = []
        for j in range(markclasscount):
            baseanchorofst = fontfile.readuint16()
            ptr = fontfile.tell()
            if baseanchorofst == 0:
                anchortables.append(None)
            else:
                anchortables.append(read_anchortable(
                    ofst+baseanchorofst, fontfile))
            fontfile.seek(ptr)
        basearray.append(anchortables)
    return basearray


class MarkToBaseSubtable(GposSubtable):
    ''' Mark-To-Base Positioning Table (GPOS Lookup Type 4) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

        self.fmt = self.fontfile.readuint16(self.ofst)
        assert self.fmt == 1
        markcovofst = self.fontfile.readuint16()
        basecovofst = self.fontfile.readuint16()
        markclasscnt = self.fontfile.readuint16()
        markarrayofst = self.fontfile.readuint16()
        basearrayofst = self.fontfile.readuint16()

        self.markarray = read_markarray_table(self.ofst+markarrayofst, self.fontfile)
        self.basearray = read_basearray(self.ofst+basearrayofst, self.fontfile, markclasscnt)
        self.markcoverage = Coverage(self.ofst+markcovofst, self.fontfile)
        self.basecoverage = Coverage(self.ofst+basecovofst, self.fontfile)

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        deltas: list[PositionDelta] = [PositionDelta(0,0,0)]
        for i, (gid, gid0) in enumerate(zip(glyphids[1:], glyphids[:-1])):
            markid = self.markcoverage.covidx(gid)
            baseid = self.basecoverage.covidx(gid0)
            if markid is None or baseid is None:
                deltas.append(PositionDelta(0, 0, 0))
            else:
                markanchor = self.markarray[markid]
                baseanchor = self.basearray[baseid][markanchor.markclass]
                if baseanchor is None:
                    deltas.append(PositionDelta(0, 0, 0))
                else:
                    dx = baseanchor.x - markanchor.anchortable.x - advances[i]
                    dy = baseanchor.y - markanchor.anchortable.y
                    deltas.append(PositionDelta(dx, dy, -dx))
                    logging.debug('Positioning Mark %s on Base %s: (%s, %s)',
                                  gid, gid0, dx, dy)

        return deltas


class MarkToMarkSubtable(GposSubtable):
    ''' Mark-To-Mark Positioning Table (GPOS Lookup Type 6) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

        self.fmt = self.fontfile.readuint16(self.ofst)
        assert self.fmt == 1
        mark1covofst = self.fontfile.readuint16()
        mark2covofst = self.fontfile.readuint16()
        markclasscnt = self.fontfile.readuint16()
        mark1arrayofst = self.fontfile.readuint16()
        mark2arrayofst = self.fontfile.readuint16()

        mark2count = self.fontfile.readuint16(self.ofst+mark2arrayofst)
        self.mark2array = []
        for i in range(mark2count):
            anchorofsts = []
            for j in range(markclasscnt):
                anchorofsts.append(self.fontfile.readuint16())
            anchortables = []
            ptr = self.fontfile.tell()
            for anchorofst in anchorofsts:
                anchortables.append(read_anchortable(
                    self.ofst+mark2arrayofst+anchorofst,
                    self.fontfile))
            self.mark2array.append(anchortables)
            self.mark1array = read_markarray_table(
                self.ofst+mark1arrayofst, self.fontfile)
            self.fontfile.seek(ptr)

        self.mark1coverage = Coverage(
            self.ofst+mark1covofst, self.fontfile)
        self.mark2coverage = Coverage(
            self.ofst+mark2covofst, self.fontfile)

    def adjust(self, *glyphids: int, advances: list[int], lookups: list[GposLookup]) -> list[PositionDelta]:
        ''' Get dx, dy, dxadvance for the glyph '''
        deltas: list[PositionDelta] = [PositionDelta(0,0,0)]
        for i, (gid, gid0) in enumerate(zip(glyphids[1:], glyphids[:-1])):
            mark1id = self.mark1coverage.covidx(gid)
            mark2id = self.mark2coverage.covidx(gid0)
            if mark1id is None or mark2id is None:
                deltas.append(PositionDelta(0, 0, 0))
            else:
                mark1anchor = self.mark1array[mark1id]
                mark2anchor = self.mark2array[mark2id][mark1anchor.markclass]
                if mark2anchor is None:
                    deltas.append(PositionDelta(0, 0, 0))
                else:
                    dx = mark2anchor.x - mark1anchor.anchortable.x - advances[i]
                    dy = mark2anchor.y - mark1anchor.anchortable.y
                    deltas.append(PositionDelta(dx, dy, -dx))
                    logging.debug('Positioning Mark %s on Mark %s: (%s, %s)',
                                  gid, gid0, dx, dy)

        return deltas
